import re

from typing import Dict, List, Set
from pathlib import Path
from neologism import DCFG, Rule

from syslog_ng_cfg_helper.driver_db import Driver, DriverDB, Block, Option
from syslog_ng_cfg_helper.globals import EXCLUSIVE_PLUGINS, PLUGIN_CONTEXTS, TYPES
from .parse_sentence import parse_sentence, ParseError


class GrammarFileMissingError(Exception):
    pass


def __find_grammar_files(driver_source_dir: Path) -> Set[Path]:
    grammar_files = set(driver_source_dir.rglob("*-grammar.y"))

    if len(grammar_files) == 0:
        raise GrammarFileMissingError()

    return grammar_files


def __format_types(grammar: DCFG) -> None:
    for type_symbol, formatted_type in TYPES:
        if type_symbol in grammar.symbols:
            grammar.make_symbol_terminal(type_symbol)
            grammar.add_rule(Rule(type_symbol, (formatted_type,)))


def __remove_ifdef(grammar: DCFG) -> None:
    for symbol in {"KW_IFDEF", "KW_ENDIF"}:
        if symbol in grammar.terminals:
            grammar.remove_symbol(symbol)


def __get_token_resolutions(parser_file: Path) -> Dict[str, Set[str]]:
    resolutions: Dict[str, Set[str]] = {}

    struct_regex = re.compile(r"CfgLexerKeyword(.*?)};")
    entry_regex = re.compile(r"{[^{}]+,[^{}]+}")

    with parser_file.open("r") as file:
        for struct_match in struct_regex.finditer(file.read().replace("\n", "")):
            for entry_match in entry_regex.finditer(struct_match.group(0)):
                entry = entry_match.group(0)[1:-1].replace(" ", "").split(",")
                token = entry[1]
                keyword = entry[0][1:-1].replace("_", "-")
                resolutions.setdefault(token, set()).add(keyword)

    return resolutions


def __resolve_tokens_to_keywords(grammar: DCFG, common_parser_file: Path, parser_file: Path) -> None:
    parser_files: List[Path] = [common_parser_file, parser_file]
    for file in parser_files:
        for token, resolutions in __get_token_resolutions(file).items():
            for resolution in resolutions:
                grammar.add_rule(Rule(token, (resolution,)))


def __prepare_module_grammar(module_source_dir: Path, common_parser_file: Path) -> DCFG:
    module_grammar = DCFG()

    for grammar_file in __find_grammar_files(module_source_dir):
        parser_file = Path(str(grammar_file).replace("-grammar.y", "-parser.c"))

        grammar = DCFG()
        grammar.load_yacc_file(str(grammar_file))

        __format_types(grammar)
        __remove_ifdef(grammar)
        __resolve_tokens_to_keywords(grammar, common_parser_file, parser_file)

        module_grammar.load_dcfg(grammar)
        module_grammar.start_symbol = grammar.start_symbol

    return module_grammar


def __merge_blocks_and_options_with_the_same_name(driver_db: DriverDB) -> None:
    def process(block: Block):
        for option in list(block.options):
            try:
                if option.name is None:
                    continue
                inner_block_with_same_name = block.get_block(option.name)
            except KeyError:
                continue

            for params in option.params:
                inner_block_with_same_name.add_option(Option(params={params}))

            block.remove_option(option.name)

        for inner_block in block.blocks:
            process(inner_block)

    for ctx in driver_db.contexts:
        for driver in driver_db.get_drivers_in_context(ctx):
            process(driver)


def __accepts_plugins(driver: Driver) -> bool:
    try:
        positional_option = driver.get_option(None)
    except KeyError:
        return False

    if not ("<plugin>",) in positional_option.params:
        return False

    return True


def __remove_plugin_param_from_driver(driver: Driver) -> None:
    new_params = set(driver.get_option(None).params)
    new_params.remove(("<plugin>",))
    driver.remove_option(None)
    if len(new_params) > 0:
        driver.add_option(Option(params=new_params))


def __connect_inner_plugins(driver_db: DriverDB) -> None:
    for plugin_context, driver_context in PLUGIN_CONTEXTS.items():
        if driver_context not in driver_db.contexts or plugin_context not in driver_db.contexts:
            continue

        plugins = driver_db.get_drivers_in_context(plugin_context)
        drivers = driver_db.get_drivers_in_context(driver_context)

        for driver in drivers:
            if not __accepts_plugins(driver):
                continue

            __remove_plugin_param_from_driver(driver)

            for plugin in plugins:
                if plugin.name in EXCLUSIVE_PLUGINS.keys() and driver.name not in EXCLUSIVE_PLUGINS[plugin.name]:
                    continue

                driver.add_block(plugin.to_block())

        driver_db.remove_context(plugin_context)


def __post_process_driver_db(driver_db: DriverDB) -> None:
    __merge_blocks_and_options_with_the_same_name(driver_db)
    __connect_inner_plugins(driver_db)


def __load_drivers_in_module(module_source_dir: Path, common_parser_file: Path) -> DriverDB:
    drivers = DriverDB()

    try:
        grammar = __prepare_module_grammar(module_source_dir, common_parser_file)
    except GrammarFileMissingError:
        print("    Skipping module: Grammar file is missing.")
        return DriverDB()

    for sentence in grammar.sentences:
        try:
            driver_slice = parse_sentence(sentence)
            drivers.add_driver(driver_slice)
        except ParseError as exception:
            print(f"    Cannot parse sentence '{' '.join(sentence)}': {exception}")

    return drivers


def load_modules(lib_dir: Path, modules_dir: Path) -> DriverDB:
    common_parser_file = lib_dir / "cfg-parser.c"
    driver_db = DriverDB()
    module_source_dirs: List[Path] = list(filter(lambda path: path.is_dir(), modules_dir.glob("*")))

    for module_source_dir in module_source_dirs:
        print(f"Loading module '{module_source_dir.name}'.")

        drivers = __load_drivers_in_module(module_source_dir, common_parser_file)
        driver_db.merge(drivers)

    __post_process_driver_db(driver_db)

    return driver_db



from typing import Iterable

__INDENTATION = 4


def indent(string: str) -> str:
    lines = string.split("\n")
    indented_lines = [f"{' ' * __INDENTATION}{line}" for line in lines]
    return "\n".join(indented_lines)


def sorted_with_none(iterable: Iterable):
    def none_to_empty_str(elem):
        if elem is None:
            return ""
        return elem

    sorted_iterable = list(iterable)
    sorted_iterable.sort(key=none_to_empty_str)

    return sorted_iterable



from __future__ import annotations
from typing import Any, Dict, Optional, ValuesView

from .exceptions import MergeException
from .option import Option
from .utils import indent, sorted_with_none


class Block:
    def __init__(self, name: str):
        self.__name = name
        self.__blocks: Dict[str, Block] = {}
        self.__options: Dict[Optional[str], Option] = {}

    @property
    def name(self) -> str:
        return self.__name

    @property
    def blocks(self) -> ValuesView[Block]:
        return self.__blocks.values()

    def get_block(self, name: str) -> Block:
        return self.__blocks[name]

    def add_block(self, block: Block) -> None:
        if block.name not in self.__blocks.keys():
            self.__blocks[block.name] = block.copy()
        else:
            self.get_block(block.name).merge(block)

    def remove_block(self, name) -> None:
        self.__blocks.pop(name)

    @property
    def options(self) -> ValuesView[Option]:
        return self.__options.values()

    def get_option(self, name: Optional[str]) -> Option:
        return self.__options[name]

    def add_option(self, option: Option) -> None:
        if option.name not in self.__options.keys():
            self.__options[option.name] = option.copy()
        else:
            self.get_option(option.name).merge(option)

    def remove_option(self, name: Optional[str]) -> None:
        self.__options.pop(name)

    def merge(self, other: Block) -> None:
        if self.name != other.name:
            raise MergeException(f"Cannot merge two Blocks with different names: '{self.name}' and '{other.name}'")

        for block in other.blocks:
            self.add_block(block)

        for option in other.options:
            self.add_option(option)

    def copy(self) -> Block:
        clone = Block(self.name)
        clone.merge(self)

        return clone

    @staticmethod
    def from_dict(as_dict: Dict[str, Any]) -> Block:
        self = Block(as_dict["name"])

        for block in as_dict["blocks"].values():
            self.add_block(Block.from_dict(block))

        for option in as_dict["options"].values():
            self.add_option(Option.from_dict(option))

        return self

    def to_dict(self) -> Dict[str, Any]:
        as_dict: Dict[str, Any] = {
            "name": self.name,
            "blocks": {},
            "options": {},
        }

        for block_name, block in self.__blocks.items():
            as_dict["blocks"][block_name] = block.to_dict()

        for option_name, option in self.__options.items():
            as_dict["options"][option_name or ""] = option.to_dict()

        return as_dict

    def __repr__(self) -> str:
        return f"Block({repr(self.__name)}, {repr(self.__blocks)}, {repr(self.__options)})"

    def __str__(self) -> str:
        string = f"{self.name}(\n"

        block_and_option_strs: Dict[Optional[str], str] = {}

        for block_name in sorted(self.__blocks.keys()):
            block_and_option_strs.setdefault(block_name, "")
            block_and_option_strs[block_name] += f"{indent(str(self.get_block(block_name)))}\n"

        for option_name in sorted_with_none(self.__options.keys()):
            block_and_option_strs.setdefault(option_name, "")
            block_and_option_strs[option_name] += f"{indent(str(self.get_option(option_name)))}\n"

        for block_or_option_name in sorted_with_none(block_and_option_strs.keys()):
            string += block_and_option_strs[block_or_option_name]

        string += ")"

        return string

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Block):
            return False

        return self.__name == other.__name and self.__blocks == other.__blocks and self.__options == other.__options



from typing import Set, Tuple
from syslog_ng_cfg_helper.driver_db import Block, Driver, Option


class ParseError(Exception):
    pass


def __is_type(symbol: str) -> bool:
    return symbol.startswith("<") and symbol.endswith(">")


def __is_block(sentence: Tuple[str, ...]) -> bool:
    if len(sentence) < 3:
        return False

    if __is_type(sentence[0]):
        return False

    if sentence[1] != "(":
        return False

    for symbol in sentence[2:]:
        if symbol == ")":
            return False

        if symbol == "(":
            return True

    raise ParseError(f"Block is not closed: {sentence}")


def __is_named_option(sentence: Tuple[str, ...]) -> bool:
    if len(sentence) < 3:
        return False

    if sentence[1] != "(":
        return False

    return True


def __is_positional_option(sentence: Tuple[str, ...]) -> bool:
    if len(sentence) < 1:
        raise ParseError(f"Too short sentence: {sentence}")

    return __is_type(sentence[0])


def __get_block_content(sentence: Tuple[str, ...]) -> Tuple[str, ...]:
    expected_number_of_right_braces = 0
    number_of_right_braces = 0

    for i, symbol in enumerate(sentence):
        if symbol == "(":
            expected_number_of_right_braces += 1

        if symbol == ")":
            number_of_right_braces += 1

        if number_of_right_braces == expected_number_of_right_braces and expected_number_of_right_braces != 0:
            return sentence[2:i]

    raise ParseError(f"Block is not closed: {sentence}")


def __get_named_option_name(sentence: Tuple[str, ...]) -> str:
    return sentence[0]


def __get_named_option_params(sentence: Tuple[str, ...]) -> Tuple[str, ...]:
    return sentence[2 : sentence.index(")")]


def __parse_block(sentence: Tuple[str, ...]) -> Tuple[Block, int]:
    block = Block(sentence[0])

    block_content = __get_block_content(sentence)
    __parse_options_in_block(block_content, block)

    number_of_processed_symbols = len(block_content) + 3
    return (block, number_of_processed_symbols)


def __parse_named_option(sentence: Tuple[str, ...]) -> Tuple[Option, int]:
    option_name = __get_named_option_name(sentence)
    option_params = __get_named_option_params(sentence)
    number_of_processed_symbols = len(option_params) + 3

    if len(option_params) == 0:
        option_params = ("<empty>",)

    option = Option(option_name, {option_params})

    return (option, number_of_processed_symbols)


def __parse_positional_option(sentence: Tuple[str, ...]) -> Tuple[Option, int]:
    return (Option(params={(sentence[0],)}), 1)


def __mark_n_symbols_as_processed(processed: Set[int], start_index: int, number_of_symbols: int) -> None:
    processed |= set(range(start_index, start_index + number_of_symbols))


def __parse_options_in_block(sentence: Tuple[str, ...], target_block: Block) -> None:
    processed: Set[int] = set()

    for i in range(len(sentence)):
        if i in processed:
            continue

        rest_of_sentence = sentence[i:]
        if __is_block(rest_of_sentence):
            block, number_of_parsed_symbols = __parse_block(rest_of_sentence)
            target_block.add_block(block)
        elif __is_named_option(rest_of_sentence):
            named_option, number_of_parsed_symbols = __parse_named_option(rest_of_sentence)
            target_block.add_option(named_option)
        elif __is_positional_option(rest_of_sentence):
            positional_option, number_of_parsed_symbols = __parse_positional_option(rest_of_sentence)
            target_block.add_option(positional_option)

        __mark_n_symbols_as_processed(processed, i, number_of_parsed_symbols)


def parse_sentence(sentence: Tuple[str, ...]) -> Driver:
    if len(sentence) < 4:
        raise ParseError("Too short sentence.")

    if not sentence[0].startswith("LL_CONTEXT_"):
        raise ParseError("Context is missing.")

    if not (sentence[2] == "(" and sentence[-1] == ")"):
        raise ParseError("Braces are missing, probably not a driver.")

    context = sentence[0].replace("LL_CONTEXT_", "").replace("_", "-").lower()
    name = sentence[1]

    driver = Driver(context, name)

    __parse_options_in_block(sentence[3:-1], driver)

    return driver



from __future__ import annotations

from typing import Dict, Any

from .exceptions import MergeException
from .block import Block
from .option import Option


class Driver(Block):
    def __init__(self, context: str, name: str) -> None:
        self.__context = context
        super().__init__(name)

    @property
    def context(self) -> str:
        return self.__context

    def copy(self) -> Driver:
        copied = Driver(self.context, self.name)
        copied.merge(self)

        return copied

    def merge(self, other: Block) -> None:
        if isinstance(other, Driver) and self.context != other.context:
            raise MergeException(
                f"Cannot merge two drivers with different contexts: '{self.context}' and '{other.context}'"
            )

        super().merge(other)

    def to_block(self) -> Block:
        block = Block(self.name)
        block.merge(self)

        return block

    @staticmethod
    def from_dict(as_dict: Dict[str, Any]) -> Driver:
        self = Driver(as_dict["context"], as_dict["name"])

        for block in as_dict["blocks"].values():
            self.add_block(Block.from_dict(block))

        for option in as_dict["options"].values():
            self.add_option(Option.from_dict(option))

        return self

    def to_dict(self) -> Dict[str, Any]:
        as_dict = super().to_dict()
        as_dict.update({"context": self.context})

        return as_dict

    def __repr__(self) -> str:
        block_repr = super().__repr__()
        return f"Driver({repr(self.context)}, {block_repr[len('Block(') :]}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Driver):
            return False

        return self.context == other.context and super().__eq__(other)



from __future__ import annotations

import json

from typing import Any, Dict, IO, KeysView, ValuesView

from .driver import Driver


class DriverDB:
    def __init__(self) -> None:
        self.__contexts: Dict[str, Dict[str, Driver]] = {}

    @property
    def contexts(self) -> KeysView[str]:
        return self.__contexts.keys()

    def add_driver(self, driver: Driver) -> DriverDB:
        context = self.__contexts.setdefault(driver.context, {})

        if driver.name in context.keys():
            context[driver.name].merge(driver)
        else:
            context[driver.name] = driver.copy()

        return self

    def get_driver(self, context: str, driver_name: str) -> Driver:
        return self.__contexts[context][driver_name]

    def get_drivers_in_context(self, context: str) -> ValuesView[Driver]:
        return self.__contexts[context].values()

    def remove_context(self, context: str) -> None:
        self.__contexts.pop(context)

    def merge(self, other: DriverDB) -> DriverDB:
        for context in other.contexts:
            for driver in other.get_drivers_in_context(context):
                self.add_driver(driver)

        return self

    @staticmethod
    def from_dict(as_dict: Dict[str, Any]) -> DriverDB:
        self = DriverDB()

        for _, drivers in as_dict["contexts"].items():
            for _, driver in drivers.items():
                self.add_driver(Driver.from_dict(driver))

        return self

    def to_dict(self) -> Dict[str, Any]:
        as_dict: Dict[str, Any] = {"contexts": {}}

        for context_name, drivers in self.__contexts.items():
            context = as_dict["contexts"].setdefault(context_name, {})

            for driver_name, driver in drivers.items():
                context[driver_name] = driver.to_dict()

        return as_dict

    @staticmethod
    def load(file: IO) -> DriverDB:
        return DriverDB.from_dict(json.load(file))

    def dump(self, file: IO) -> None:
        json.dump(self.to_dict(), file)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DriverDB):
            return False

        return self.__contexts == other.__contexts

    def __repr__(self) -> str:
        return f"DriverDB({repr(self.__contexts)})"
